%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This script is a supplementary material to the Paper:
%              Title: Encoding Color Difference Signals for High Dynamic Range and Wide Gamut Imagery
%            Authors: Jan Froehlich, Timo Kunkel, Robin Atkins, Jaclyn Pytlarz, Scott Daly, Andreas Schilling, Bernd Eberhardt
%         Conference: 23th IS&T Color and Imaging Conference, 2015
% Conference Website: http://www.ingentaconnect.com/contentone/ist/cic/2015/00002015/00000001/art00044
%      PDF Available: https://www.hdm-stuttgart.de/~froehlichj/
%
%
clear all
close all
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Prep XYZ to ICaCb conversion
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Matrix to convert from XYZ to the ICaCb primaries
M1__XYZ_2_RGBoverline = [ 0.37613 0.70431 -0.05675;...
                         -0.21649 1.14744  0.05356;...
                          0.02567 0.16713  0.74235];

% SMPTE 2084 nonlinearity (PQ)
n = 2610/16384; m = 2523/32; c1 = 3424/4096; c2 = 2413/128; c3 = 2392/128;
nonlinearity__linear_2_PQ = @(x)((c1+c2*(x./10000).^n) ./ (1+c3*(x./10000).^n)).^m;
clear n m c1 c2 c3

% Decorrelation Matrix
M2__RGBoverlinedash_2_ICaCb = [0.4949  0.5037 0.0015;...
                               4.2854 -4.5462 0.2609;...
                               0.3605 1.1499 -1.5105];

% These functions expect RGB pixels with dim1 = number of pixels and R, G, B in dim2. 
XYZ_2_ICaCb = @(x)nonlinearity__linear_2_PQ(x * M1__XYZ_2_RGBoverline') * M2__RGBoverlinedash_2_ICaCb';

%% Check against some example values:
black = [0 0 0];
XYZ_2_ICaCb(black) % should result in values below 10^-5

D65_100cdm2 = [95.0456  100.0000  108.9058];
XYZ_2_ICaCb(D65_100cdm2) % should result in ~ [0.5081 0 0]

D65_10000cdm2 = [9504.55927 10000 10890.5775];
XYZ_2_ICaCb(D65_10000cdm2) % should result in ~ [1 0 0]

sRGB100red = [41.2391   21.2639    1.9331];
XYZ_2_ICaCb(sRGB100red) % should result in ~ [0.3645 0.2304 0.1331]


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Visualize some common gamuts
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Create Gamut hull
precision = 64;
grid2D = [];
[grid2D(:,:,2), grid2D(:,:,1)] = meshgrid(linspace(0,1,precision),linspace(0,1,precision));
g2D = reshape(grid2D(1:1:(precision),1:1:(precision),:),[],2);
o = ones((precision)^2,1); z = zeros((precision)^2,1);
gamutHull = cat(1, cat(2,g2D,z), cat(2,1-g2D,o), cat(1,cat(2,1-g2D,z),cat(2,g2D,o))*[0 0 1;0 1 0;1 0 0],...
    cat(1,cat(2,(repmat([0 1],(precision)^2,1)-g2D)*[-1 0; 0 1],z),...
    cat(2,(repmat([1 0],(precision)^2,1)-g2D)*[1 0; 0 -1],o))*[1 0 0;0 0 1;0 1 0]);

%% Visualize sRGB (Theoretical with black = 0 and white = 100cd/m2)
sRGB_2_linear = @(x)((x>0.04045).*((x+0.055)./1.055).^2.4 + (x<=0.04045).*(x./12.92));
sRGB_2_XYZ = [0.4124 0.3576 0.1805; 0.2126 0.7152 0.0722; 0.0193 0.1192 0.9505];

% Linearize, convert to XYZ and sibsequently to ICaCb
gamutHull_ICaCb = XYZ_2_ICaCb(sRGB_2_linear(gamutHull)*sRGB_2_XYZ'*100);

%% Plot
figure('Name','sRGB in ICaCb')
scatter3(gamutHull_ICaCb(:,1), gamutHull_ICaCb(:,2), gamutHull_ICaCb(:,3), 10, gamutHull, 'filled');
xlabel('I'); ylabel('C_a'); zlabel('C_b');
xlim([0 1]); ylim([-0.5 0.5]); zlim([-0.5 0.5]);
title('sRGB in ICaCb')
grid on
%% View Ca/Cb plane
campos([1 0 0]), camtarget([0 0 0]), camup([0 0 1])
%% View I/Ca plane
campos([0.5 0 1]), camtarget([0.5 0 0]), camup([1 0 0])
%% View I/Cb plane
campos([0.5 -1 0]), camtarget([0.5 0 0]), camup([1 0 0])


%% Visualize Theoretical Rec.2020 HDR display 0.005...10000cd/m2, gamma 2.8
Rec2020_2_XYZ = [0.6370 0.1446 0.1689; 0.2627 0.6780 0.0593; 0.0000 0.0281 1.0610];

% Linearize, convert to XYZ and subsequently to ICaCb
gamutHull_ICaCb = XYZ_2_ICaCb((gamutHull.^2.8)*Rec2020_2_XYZ'*10000);

%% Plot
figure('Name','Rec.2020 10000cd/m2 peak white in ICaCb')
scatter3(gamutHull_ICaCb(:,1), gamutHull_ICaCb(:,2), gamutHull_ICaCb(:,3), 14, gamutHull, 'filled');
xlabel('I'); ylabel('C_a'); zlabel('C_b');
xlim([0 1]); ylim([-0.5 0.5]); zlim([-0.5 0.5]);
title('Rec.2020 10000cd/m2 peak white in ICaCb')
grid on
%% View Ca/Cb plane
campos([1 0 0]), camtarget([0 0 0]), camup([0 0 1])
%% View I/Ca plane
campos([0.5 0 1]), camtarget([0.5 0 0]), camup([1 0 0])
%% View I/Cb plane
campos([0.5 -1 0]), camtarget([0.5 0 0]), camup([1 0 0])

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Convert an SDR image
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
img_sRGB = double(imread('peppers.png'))./255;
imshow(img_sRGB)

% linearize and convert to to XYZ, assuming an sRGB monitor with 100cd/m2 peak white and perfect black 
sRGB_2_linear = @(x)((x>0.04045).*((x+0.055)./1.055).^2.4 + (x<=0.04045).*(x./12.92));
sRGB_2_XYZ = [0.4124 0.3576 0.1805; 0.2126 0.7152 0.0722; 0.0193 0.1192 0.9505];

img_linear_sRGB = sRGB_2_linear(img_sRGB).*100;

% reshape 3d array to 2d line array
imgLine_linear_SRGB = reshape( img_linear_sRGB , [] , 3);

% convert to XYZ
imgLine_linear_XYZ = imgLine_linear_SRGB * sRGB_2_XYZ';
   
% convert to ICaCb
imgLine_ICaCb = XYZ_2_ICaCb(imgLine_linear_XYZ);

% reshape back from 2d line array to 3d array
img_ICaCb = reshape( imgLine_ICaCb, size(img_linear_sRGB,1), size(img_linear_sRGB,2), 3);

% show the three RGB channels
figure('units','normalized','outerposition',[0 0 1 1])
subplot(3,4,1);
imshow(img_sRGB(:,:,1))
title('R')
subplot(3,4,5);
imshow(img_sRGB(:,:,2))
title('G')
subplot(3,4,9);
imshow(img_sRGB(:,:,3))
title('B')

% Check corellation
subplot(3,4,2);
sp = scatter(reshape(img_sRGB(:,:,1),[],1), reshape(img_sRGB(:,:,2),[],1), '.k');
sp.MarkerEdgeAlpha = 0.01; sp.MarkerFaceAlpha = 0.01;
xlabel('R'); ylabel('G'); grid on
subplot(3,4,6);
sp = scatter(reshape(img_sRGB(:,:,2),[],1), reshape(img_sRGB(:,:,3),[],1), '.k');
sp.MarkerEdgeAlpha = 0.01; sp.MarkerFaceAlpha = 0.01;
xlabel('G'); ylabel('B'); grid on
subplot(3,4,10);
sp = scatter(reshape(img_sRGB(:,:,3),[],1), reshape(img_sRGB(:,:,1),[],1), '.k');
sp.MarkerEdgeAlpha = 0.01; sp.MarkerFaceAlpha = 0.01;
xlabel('B'); ylabel('R'); grid on

% show the three ICaCb channels
subplot(3,4,3);
imshow(img_ICaCb(:,:,1))
title('I')
subplot(3,4,7);
imshow(img_ICaCb(:,:,2) + 0.5)
title('C_a + 0.5')
subplot(3,4,11);
imshow(img_ICaCb(:,:,3) + 0.5)
title('B_b + 0.5')

% Check corellation
subplot(3,4,4);
sp = scatter(imgLine_ICaCb(:,2), imgLine_ICaCb(:,1), '.k');
sp.MarkerEdgeAlpha = 0.01; sp.MarkerFaceAlpha = 0.01;
xlabel('C_a'); ylabel('I'); grid on
subplot(3,4,8);
sp = scatter(imgLine_ICaCb(:,3), imgLine_ICaCb(:,1), '.k');
sp.MarkerEdgeAlpha = 0.01; sp.MarkerFaceAlpha = 0.01;
xlabel('C_b'); ylabel('I'); grid on
subplot(3,4,12);
sp = scatter(imgLine_ICaCb(:,2), imgLine_ICaCb(:,3), '.k');
sp.MarkerEdgeAlpha = 0.01; sp.MarkerFaceAlpha = 0.01;
xlabel('C_a'); ylabel('C_b'); grid on

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Convert an HDR image
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
img_Rec2020_PQ = double(imread('ShowGirl2TeaserClip4000_1920x1080p_25_12_P3_ct2020_444i_060.tif'))./65535;

% linearize and convert to to XYZ, assuming an sRGB monitor with 100cd/m2 peak white and perfect black
n = 2610/16384; m = 2523/32; c1 = 3424/4096; c2 = 2413/128; c3 = 2392/128;
PQ_2_Linear = @(x)(((x.^(1/m)-c1) ./ (c2-c3*x.^(1/m))).^(1/n))*10000;
clear n m c1 c2 c3

Rec2020_2_XYZ = [0.6370 0.1446 0.1689; 0.2627 0.6780 0.0593; 0.0000 0.0281 1.0610];

img_linear_Rec2020 = PQ_2_Linear(img_Rec2020_PQ);

% reshape 3d array to 2d line array
imgLine_linear_Rec2020 = reshape( img_linear_Rec2020 , [] , 3);

% convert to XYZ
imgLine_linear_XYZ = imgLine_linear_Rec2020 * Rec2020_2_XYZ';
   
% convert to ICaCb
imgLine_ICaCb = XYZ_2_ICaCb(imgLine_linear_XYZ);

% reshape back from 2d line array to 3d array
img_ICaCb = reshape( imgLine_ICaCb, size(img_linear_Rec2020,1), size(img_linear_Rec2020,2), 3);

% show the three RGB channels
figure('units','normalized','outerposition',[0 0 1 1])
subplot(3,4,1);
imshow(img_Rec2020_PQ(:,:,1))
title('R')
subplot(3,4,5);
imshow(img_Rec2020_PQ(:,:,2))
title('G')
subplot(3,4,9);
imshow(img_Rec2020_PQ(:,:,3))
title('B')

% Check corellation
subplot(3,4,2);
sp = scatter(reshape(img_Rec2020_PQ(:,:,1),[],1), reshape(img_Rec2020_PQ(:,:,2),[],1), '.k');
sp.MarkerEdgeAlpha = 0.01; sp.MarkerFaceAlpha = 0.01;
xlabel('R'); ylabel('G'); grid on
subplot(3,4,6);
sp = scatter(reshape(img_Rec2020_PQ(:,:,2),[],1), reshape(img_Rec2020_PQ(:,:,3),[],1), '.k');
sp.MarkerEdgeAlpha = 0.01; sp.MarkerFaceAlpha = 0.01;
xlabel('G'); ylabel('B'); grid on
subplot(3,4,10);
sp = scatter(reshape(img_Rec2020_PQ(:,:,3),[],1), reshape(img_Rec2020_PQ(:,:,1),[],1), '.k');
sp.MarkerEdgeAlpha = 0.01; sp.MarkerFaceAlpha = 0.01;
xlabel('B'); ylabel('R'); grid on

% show the three ICaCb channels
subplot(3,4,3);
imshow(img_ICaCb(:,:,1))
title('I')
subplot(3,4,7);
imshow(img_ICaCb(:,:,2) + 0.5)
title('C_a + 0.5')
subplot(3,4,11);
imshow(img_ICaCb(:,:,3) + 0.5)
title('B_b + 0.5')

% Check corellation
subplot(3,4,4);
sp = scatter(imgLine_ICaCb(:,2), imgLine_ICaCb(:,1), '.k');
sp.MarkerEdgeAlpha = 0.01; sp.MarkerFaceAlpha = 0.01;
xlabel('C_a'); ylabel('I'); grid on
subplot(3,4,8);
sp = scatter(imgLine_ICaCb(:,3), imgLine_ICaCb(:,1), '.k');
sp.MarkerEdgeAlpha = 0.01; sp.MarkerFaceAlpha = 0.01;
xlabel('C_b'); ylabel('I'); grid on
subplot(3,4,12);
sp = scatter(imgLine_ICaCb(:,2), imgLine_ICaCb(:,3), '.k');
sp.MarkerEdgeAlpha = 0.01; sp.MarkerFaceAlpha = 0.01;
xlabel('C_a'); ylabel('C_b'); grid on





